"""
Auxiliary calculations needed for the tricrypto implementation.

This is loosely the python counterpart to the Tricrypto math contract.
"""
from math import isqrt
from typing import List

from gmpy2 import mpz

from curvesim.exceptions import CalculationError, CurvesimValueError
from curvesim.logging import get_logger

logger = get_logger(__name__)

NOISE_FEE = 10**5  # 0.1 bps

MIN_GAMMA = 10**10
MAX_GAMMA = 5 * 10**16

EXP_PRECISION = 10**10

PRECISION = 10**18  # The precision to convert to
A_MULTIPLIER = 10000

N_COINS = 3
MIN_A: int = N_COINS**N_COINS * A_MULTIPLIER // 100
MAX_A: int = N_COINS**N_COINS * A_MULTIPLIER * 1000


# pylint: disable-next=too-many-locals,too-many-statements,too-many-branches
def get_y(  # noqa: complexity: 18
    ANN: int,
    gamma: int,
    x: List[int],
    D: int,
    i: int,
) -> List[int]:
    """
    @notice Calculate x[i] given other balances x[0..N_COINS-1] and invariant D.
    @dev ANN = A * N**N.
    @param _ANN AMM.A() value.
    @param _gamma AMM.gamma() value.
    @param x Balances multiplied by prices and precisions of all coins.
    @param _D Invariant.
    @param i Index of coin to calculate y.
    """
    # Safety checks
    assert MIN_A <= ANN <= MAX_A
    assert MIN_GAMMA <= gamma <= MAX_GAMMA
    assert 10**17 <= D <= 10**15 * 10**18

    frac: int = 0
    for k in range(3):
        if k != i:
            frac = x[k] * 10**18 // D
            assert 10**16 <= frac <= 10**20, "Unsafe values x[i]"
            # if above conditions are met, x[k] > 0

    j: int = 0
    k: int = 0
    if i == 0:
        j = 1
        k = 2
    elif i == 1:
        j = 0
        k = 2
    elif i == 2:
        j = 0
        k = 1

    x_j: int = mpz(x[j])
    x_k: int = mpz(x[k])
    gamma = mpz(gamma)
    gamma2: int = gamma**2

    a: int = 10**36 // 27

    # 10**36/9 + 2*10**18*gamma/27
    #   - D**2/x_j*gamma**2*ANN/27**2/convert(A_MULTIPLIER, int256)/x_k
    b: int = (
        10**36 // 9
        + 2 * 10**18 * gamma // 27
        - D**2 // x_j * gamma2 * ANN // 27**2 // A_MULTIPLIER // x_k
    )

    # Vyper does signed integer division, rounding towards zero, so we need
    # to track the sign of b, to flip signs before and after python integer
    # division.
    b_is_neg = b < 0

    # 10**36/9 + gamma*(gamma + 4*10**18)/27
    #   + gamma**2*(x_j+x_k-D)/D*ANN/27/convert(A_MULTIPLIER, int256)
    c: int = 10**36 // 9 + gamma * (gamma + 4 * 10**18) // 27
    _c_neg: int = x_j + x_k - D
    if _c_neg < 0:
        # since vyper will do signed integer division, which rounds toward 0,
        # we switch signs to round appropriately and then change the sign back
        _c: int = gamma2 * (-1 * _c_neg) // D * ANN // 27 // A_MULTIPLIER
        _c *= -1
    else:
        _c: int = gamma2 * _c_neg // D * ANN // 27 // A_MULTIPLIER
    c += _c

    # (10**18 + gamma)**2/27
    d: int = (10**18 + gamma) ** 2 // 27

    # abs(3*a*c/b - b)
    d0: int = abs(3 * a * c // b - b)  # <------------ a is smol.

    divider: int = 0
    if d0 > 10**48:
        divider = 10**30
    elif d0 > 10**44:
        divider = 10**26
    elif d0 > 10**40:
        divider = 10**22
    elif d0 > 10**36:
        divider = 10**18
    elif d0 > 10**32:
        divider = 10**14
    elif d0 > 10**28:
        divider = 10**10
    elif d0 > 10**24:
        divider = 10**6
    elif d0 > 10**20:
        divider = 10**2
    else:
        divider = 1

    additional_prec: int = 0
    if b_is_neg:
        b *= -1
    if abs(a) > abs(b):
        additional_prec = abs(a // b)
        a = a * additional_prec // divider
        b = b * additional_prec // divider
        c = c * additional_prec // divider
        d = d * additional_prec // divider
    else:
        additional_prec = abs(b // a)
        a = a // additional_prec // divider
        b = b // additional_prec // divider
        c = c // additional_prec // divider
        d = d // additional_prec // divider
    if b_is_neg:
        b *= -1

    # 3*a*c/b - b
    _3ac: int = 3 * a * c
    if b_is_neg:
        delta0: int = -(_3ac // -b) - b
    else:
        delta0: int = _3ac // b - b

    # 9*a*c/b - 2*b - 27*a**2/b*d/b
    if b_is_neg:
        delta1: int = -(3 * _3ac // -b) - 2 * b - 27 * a**2 // -b * d // -b
    else:
        delta1: int = 3 * _3ac // b - 2 * b - 27 * a**2 // b * d // b

    # delta1**2 + 4*delta0**2/b*delta0
    if b_is_neg:
        sqrt_arg: int = delta1**2 - (4 * delta0**2 // -b * delta0)
    else:
        sqrt_arg: int = delta1**2 + 4 * delta0**2 // b * delta0

    sqrt_val: int = 0
    if sqrt_arg > 0:
        sqrt_val = isqrt(sqrt_arg)
    else:
        return [_newton_y(ANN, gamma, x, D, i), 0]

    b_cbrt: int = 0
    if b >= 0:
        b_cbrt = _cbrt(b)
    else:
        b_cbrt = -_cbrt(-b)

    second_cbrt: int = 0
    if delta1 > 0:
        second_cbrt = _cbrt((delta1 + sqrt_val) // 2)
    else:
        second_cbrt = -_cbrt(-(delta1 - sqrt_val) // 2)

    # b_cbrt*b_cbrt/10**18*second_cbrt/10**18
    if second_cbrt < 0:
        C1: int = -(b_cbrt * b_cbrt // 10**18 * -second_cbrt // 10**18)
    else:
        C1: int = b_cbrt * b_cbrt // 10**18 * second_cbrt // 10**18

    # (b + b*delta0/C1 - C1)/3
    if sign(b * delta0) != sign(C1):
        root_K0: int = (b + -(b * delta0 // -C1) - C1) // 3
    else:
        root_K0: int = (b + b * delta0 // C1 - C1) // 3

    # D*D/27/x_k*D/x_j*root_K0/a
    root: int = D * D // 27 // x_k * D // x_j * root_K0 // a

    out: List[int] = [int(root), int(10**18 * root_K0 // a)]

    frac = out[0] * 10**18 // D
    assert 10**16 <= frac <= 10**20, "Unsafe value for y"
    # due to precision issues, get_y can be off by 2 wei or so wrt _newton_y

    return out


def sign(x):
    """Return +/-1 depending on sign of number.  0 is positive."""
    return -1 if x < 0 else 1


def _newton_y(  # noqa: complexity: 11  # pylint: disable=duplicate-code,too-many-locals
    ANN: int,
    gamma: int,
    x: List[int],
    D: int,
    i: int,
) -> int:
    """
    Calculating x[i] given other balances x[0..n_coins-1] and invariant D
    ANN = A * N**N
    """
    n_coins: int = len(x)

    # Safety checks
    if not MIN_A <= ANN <= MAX_A:
        raise CurvesimValueError("Unsafe value for A")
    if not MIN_GAMMA <= gamma <= MAX_GAMMA:
        raise CurvesimValueError("Unsafe value for gamma")
    assert 10**17 <= D <= 10**15 * 10**18
    for k in range(n_coins):
        if k != i:
            frac: int = x[k] * 10**18 // D
            assert 10**16 <= frac <= 10**20

    y: int = D // n_coins
    K0_i: int = 10**18
    S_i: int = 0

    x_sorted: List[int] = x.copy()
    x_sorted[i] = 0
    x_sorted = sorted(x_sorted, reverse=True)  # From high to low

    convergence_limit: int = max(max(x_sorted[0] // 10**14, D // 10**14), 100)

    for j in range(2, n_coins + 1):
        _x: int = x_sorted[n_coins - j]
        y = y * D // (_x * n_coins)  # Small _x first
        S_i += _x
    for j in range(n_coins - 1):
        K0_i = K0_i * x_sorted[j] * n_coins // D  # Large _x first

    y = mpz(y)
    K0_i = mpz(K0_i)
    D = mpz(D)
    for _ in range(255):
        y_prev: int = y

        K0: int = K0_i * y * n_coins // D
        S: int = S_i + y

        _g1k0: int = abs(gamma + 10**18 - K0) + 1

        # D / (A * N**N) * _g1k0**2 / gamma**2
        mul1: int = 10**18 * D // gamma * _g1k0 // gamma * _g1k0 * A_MULTIPLIER // ANN

        # 2*K0 / _g1k0
        mul2: int = 10**18 + (2 * 10**18) * K0 // _g1k0

        yfprime: int = 10**18 * y + S * mul2 + mul1
        _dyfprime: int = D * mul2
        if yfprime < _dyfprime:
            y = y_prev // 2
            continue

        yfprime -= _dyfprime
        fprime: int = yfprime // y

        # y -= f / f_prime;  y = (y * fprime - f) / fprime
        # y = (yfprime + 10**18 * D - 10**18 * S)
        #   / fprime + mul1 / fprime * (10**18 - K0) / K0
        y_minus: int = mul1 // fprime
        y_plus: int = (yfprime + 10**18 * D) // fprime + y_minus * 10**18 // K0
        y_minus += 10**18 * S // fprime

        if y_plus < y_minus:
            y = y_prev // 2
        else:
            y = y_plus - y_minus

        diff: int = abs(y - y_prev)
        if diff < max(convergence_limit, y // 10**14):
            frac: int = y * 10**18 // D
            assert 10**16 <= frac <= 10**20  # dev: unsafe value for y
            return int(y)

    raise CalculationError("Did not converge")


def geometric_mean(x: List[int]) -> int:
    """
    The geometric mean for 3 integers:

    (x[0] * x[1] * x[2]) ** (1/3)
    """
    prod: int = x[0] * x[1] // 10**18 * x[2] // 10**18

    if prod == 0:
        return 0

    return _cbrt(prod)


def lp_price(virtual_price, price_oracle) -> int:
    """
    Returns an LP token price approximating behavior as a constant-product AMM.
    """
    # TODO: find/implement integer cube root function
    # price_oracle = self.internal_price_oracle()
    # return (
    #     3 * self.virtual_price * icbrt(price_oracle[0] * price_oracle[1])
    # ) // 10**24
    raise CalculationError("LP price calc doesn't support more than 3 coins")


def get_p(
    xp: List[int],
    D: int,
    A: int,
    gamma: int,
) -> List[int]:
    """
    @notice Calculates dx/dy.
    @dev Output needs to be multiplied with price_scale to get the actual value.
    @param _xp Balances of the pool.
    @param _D Current value of D.
    @param _A_gamma Amplification coefficient and gamma.
    """
    assert 10**17 <= D <= 10**15 * 10**18

    N = len(xp)

    # K0 = P * N**N / D**N.
    # K0 is dimensionless and has 10**36 precision:
    K0: int = N**N * xp[0] * xp[1] // D * xp[2] // D * 10**36 // D

    # GK0 is in 10**36 precision and is dimensionless.
    # GK0 = (
    #     2 * K0 * K0 / 10**36 * K0 / 10**36
    #     + (gamma + 10**18)**2
    #     - (K0 * K0 / 10**36 * (2 * gamma + 3 * 10**18) / 10**18)
    # )
    # GK0 is always positive. So the following should never revert:
    GK0: int = (
        2 * K0**2 // 10**36 * K0 // 10**36
        + (gamma + 10**18) ** 2
        - (K0**2 // 10**36 * (2 * gamma + 3 * 10**18) // 10**18)
    )

    # NNAG2 = A * gamma**2
    NNAG2: int = A * gamma**2 // A_MULTIPLIER

    # denominator = (GK0 + NNAG2 * x / D * K0 / 10**36)
    denominator: int = GK0 + NNAG2 * xp[0] // D * K0 // 10**36

    # p_xy = x * (GK0 + NNAG2 * y / D * K0 / 10**36) / y * 10**18 / denominator
    # p_xz = x * (GK0 + NNAG2 * z / D * K0 / 10**36) / z * 10**18 / denominator
    # p is in 10**18 precision.
    return [
        (
            xp[0]
            * (GK0 + NNAG2 * xp[1] // D * K0 // 10**36)
            // xp[1]
            * 10**18
            // denominator
        ),
        (
            xp[0]
            * (GK0 + NNAG2 * xp[2] // D * K0 // 10**36)
            // xp[2]
            * 10**18
            // denominator
        ),
    ]


def newton_D(  # pylint: disable=too-many-locals
    ANN: int,
    gamma: int,
    x_unsorted: List[int],
    K0_prev: int = 0,
) -> int:
    """
    @notice Finding the invariant via newtons method using good initial guesses.
    @dev ANN is higher by the factor A_MULTIPLIER
    @dev ANN is already A * N**N
    @param ANN the A * N**N value
    @param gamma the gamma value
    @param x_unsorted the array of coin balances (not sorted)
    @param K0_prev apriori for newton's method derived from get_y_int. Defaults
           to zero (no apriori)
    """
    x: List[int] = sorted(x_unsorted, reverse=True)
    assert x[0] < (2**256 - 1) // 10**18 * N_COINS**N_COINS
    assert x[0] > 0

    S: int = sum(x)
    D: int = 0

    if K0_prev == 0:
        # Geometric mean of 3 numbers cannot be larger than the largest number
        # so the following is safe to do:
        D = N_COINS * geometric_mean(x)
    else:
        if S > 10**36:
            D = _cbrt(x[0] * x[1] // 10**36 * x[2] // K0_prev * 27 * 10**12)
        elif S > 10**24:
            D = _cbrt(x[0] * x[1] // 10**24 * x[2] // K0_prev * 27 * 10**6)
        else:
            D = _cbrt(x[0] * x[1] // 10**18 * x[2] // K0_prev * 27)

        # D not zero here if K0_prev > 0, and we checked if x[0] is gt 0.

    # initialise variables:
    K0: int = 0
    _g1k0: int = 0
    mul1: int = 0
    mul2: int = 0
    neg_fprime: int = 0
    D_plus: int = 0
    D_minus: int = 0
    D_prev: int = 0

    diff: int = 0
    frac: int = 0

    D = mpz(D)

    for _ in range(255):
        D_prev = D

        # K0 = 10**18 * x[0] * N_COINS / D * x[1] * N_COINS / D * x[2] * N_COINS / D
        K0 = 10**18 * x[0] * N_COINS // D * x[1] * N_COINS // D * x[2] * N_COINS // D
        # <-------- We can convert the entire expression using unsafe math.
        #   since x_i is not too far from D, so overflow is not expected. Also
        #      D > 0, since we proved that already. unsafe_div is safe. K0 > 0
        #        since we can safely assume that D < 10**18 * x[0]. K0 is also
        #                            in the range of 10**18 (it's a property).

        _g1k0 = gamma + 10**18  # <--------- safe to do unsafe_add.

        # The following operations can safely be unsafe.
        _g1k0: int = abs(_g1k0 - K0) + 1

        # D / (A * N**N) * _g1k0**2 / gamma**2
        # mul1 = 10**18 * D / gamma * _g1k0 / gamma * _g1k0 * A_MULTIPLIER / ANN
        mul1 = 10**18 * D // gamma * _g1k0 // gamma * _g1k0 * A_MULTIPLIER // ANN
        # <------ Since D > 0, gamma is small, _g1k0 is small, the rest are
        #        non-zero and small constants, and D has a cap in this method,
        #                    we can safely convert everything to unsafe maths.

        # 2*N*K0 / _g1k0
        # mul2 = (2 * 10**18) * N_COINS * K0 / _g1k0
        mul2 = 2 * 10**18 * N_COINS * K0 // _g1k0
        # <--------------- K0 is approximately around D, which has a cap of
        #      10**15 * 10**18 + 1, since we get that in get_y which is called
        #    with newton_D. _g1k0 > 0, so the entire expression can be unsafe.

        # neg_fprime: int = (S + S * mul2 / 10**18)
        #                   + mul1 * N_COINS / K0 - mul2 * D / 10**18
        neg_fprime = (
            (S + S * mul2 // 10**18) + mul1 * N_COINS // K0 - mul2 * D // 10**18
        )
        # <--- mul1 is a big number but not huge: safe to unsafely multiply
        # with N_coins. neg_fprime > 0 if this expression executes.
        # mul2 is in the range of 10**18, since K0 is in that range, S * mul2
        # is safe. The first three sums can be done using unsafe math safely
        # and since the final expression will be small since mul2 is small, we
        # can safely do the entire expression unsafely.

        # D -= f / fprime
        # D * (neg_fprime + S) / neg_fprime
        D_plus = D * (neg_fprime + S) // neg_fprime

        # D*D / neg_fprime
        D_minus = D * D // neg_fprime

        # Since we know K0 > 0, and neg_fprime > 0, several unsafe operations
        # are possible in the following. Also, (10**18 - K0) is safe to mul.
        # So the only expressions we keep safe are (D_minus + ...) and (D * ...)
        if 10**18 > K0:
            # D_minus += D * (mul1 / neg_fprime) / 10**18 * (10**18 - K0) / K0
            D_minus += D * (mul1 // neg_fprime) // 10**18 * (10**18 - K0) // K0
        else:
            # D_minus -= D * (mul1 / neg_fprime) / 10**18 * (K0 - 10**18) / K0
            D_minus -= (D * mul1 // neg_fprime // 10**18 * (K0 - 10**18)) // K0

        if D_plus > D_minus:
            D = D_plus - D_minus  # <--------- Safe since we check.
        else:
            D = (D_minus - D_plus) // 2

        diff = abs(D - D_prev)
        # Could reduce precision for gas efficiency here:
        if diff * 10**14 < max(10**16, D):
            # Test that we are safe with the next get_y
            for _x in x:
                frac: int = (_x * 10**18) // D
                assert 10**16 <= frac <= 10**20, "Unsafe values x[i]"
            return int(D)

    raise CalculationError("Did not converge")


def _cbrt(x: int) -> int:

    xx: int = 0
    if x >= 115792089237316195423570985008687907853269 * 10**18:
        xx = x
    elif x >= 115792089237316195423570985008687907853269:
        xx = x * 10**18
    else:
        xx = x * 10**36

    log2x: int = _snekmate_log_2(xx, False)

    # When we divide log2x by 3, the remainder is (log2x % 3).
    # So if we just multiply 2**(log2x/3) and discard the remainder to calculate our
    # guess, the newton method will need more iterations to converge to a solution,
    # since it is missing that precision. It's a few more calculations now to do less
    # calculations later:
    # pow = log2(x) // 3
    # remainder = log2(x) % 3
    # initial_guess = 2 ** pow * cbrt(2) ** remainder
    # substituting -> 2 = 1.26 ≈ 1260 / 1000, we get:
    #
    # initial_guess = 2 ** pow * 1260 ** remainder // 1000 ** remainder

    remainder: int = log2x % 3
    a: int = (2 ** (log2x // 3) * (1260**remainder)) // 1000**remainder

    # Because we chose good initial values for cube roots, 7 newton raphson iterations
    # are just about sufficient. 6 iterations would result in non-convergences, and 8
    # would be one too many iterations. Without initial values, the iteration count
    # can go up to 20 or greater. The iterations are unrolled. This reduces gas costs
    # but takes up more bytecode:
    a = (2 * a + xx // (a * a)) // 3
    a = (2 * a + xx // (a * a)) // 3
    a = (2 * a + xx // (a * a)) // 3
    a = (2 * a + xx // (a * a)) // 3
    a = (2 * a + xx // (a * a)) // 3
    a = (2 * a + xx // (a * a)) // 3
    a = (2 * a + xx // (a * a)) // 3

    if x >= 115792089237316195423570985008687907853269 * 10**18:
        a = a * 10**12
    elif x >= 115792089237316195423570985008687907853269:
        a = a * 10**6

    return a


def _snekmate_log_2(x: int, roundup: bool) -> int:
    """
    @notice An `internal` helper function that returns the log in base 2
         of `x`, following the selected rounding direction.
    @dev This implementation is derived from Snekmate, which is authored
         by pcaversaccio (Snekmate), distributed under the AGPL-3.0 license.
         https://github.com/pcaversaccio/snekmate
    @dev Note that it returns 0 if given 0. The implementation is
         inspired by OpenZeppelin's implementation here:
         https://github.com/OpenZeppelin/openzeppelin-contracts/blob/master/contracts/utils/math/Math.sol.
    @param x The 32-byte variable.
    @param roundup The Boolean variable that specifies whether
           to round up or not. The default `False` is round down.
    @return int The 32-byte calculation result.
    """
    value: int = x
    result: int = 0

    # The following lines cannot overflow because we have the well-known
    # decay behaviour of `log_2(max_value(uint256)) < max_value(uint256)`.
    if x >> 128 != 0:
        value = x >> 128
        result = 128
    if value >> 64 != 0:
        value = value >> 64
        result = result + 64
    if value >> 32 != 0:
        value = value >> 32
        result = result + 32
    if value >> 16 != 0:
        value = value >> 16
        result = result + 16
    if value >> 8 != 0:
        value = value >> 8
        result = result + 8
    if value >> 4 != 0:
        value = value >> 4
        result = result + 4
    if value >> 2 != 0:
        value = value >> 2
        result = result + 2
    if value >> 1 != 0:
        result = result + 1

    if roundup and (1 << result) < x:
        result = result + 1

    return result


def wad_exp(x: int) -> int:
    """
    @notice Calculates the e**x with 1e18 precision
    @param x The 32-byte variable to calculate the exponential of.
    @return int256 The 32-byte calculation result.

    @dev Calculates the natural exponential function of a signed integer with
         a precision of 1e18.
    @notice Note that this function consumes about 810 gas units. The implementation
            is inspired by Remco Bloemen's implementation under the MIT license here:
            https://xn--2-umb.com/22/exp-ln.
    @dev This implementation is derived from Snekmate, which is authored
         by pcaversaccio (Snekmate), distributed under the AGPL-3.0 license.
         https://github.com/pcaversaccio/snekmate
    """
    value: int = x

    # If the result is `< 0.5`, we return zero. This happens when we have the following:
    # "x <= floor(log(0.5e18) * 1e18) ~ -42e18".
    if x <= -42139678854452767551:
        return 0

    # When the result is "> (2 ** 255 - 1) / 1e18" we cannot represent it
    # as a signed integer.
    # This happens when "x >= floor(log((2 ** 255 - 1) / 1e18) * 1e18) ~ 135".
    assert x < 135305999368893231589, "wad_exp overflow"

    # `x` is now in the range "(-42, 136) * 1e18".
    # Convert to "(-42, 136) * 2 ** 96" for higher intermediate precision
    # and a binary base. This base conversion is a multiplication with
    # "1e18 / 2 ** 96 = 5 ** 18 / 2 ** 78".
    value = (x << 78) // 5**18

    # Reduce the range of `x` to "(-½ ln 2, ½ ln 2) * 2 ** 96" by
    # factoring out powers of two so that "exp(x) = exp(x') * 2 ** k",
    # where `k` is a signer integer. Solving this gives
    # "k = round(x / log(2))" and "x' = x - k * log(2)".
    # Thus, `k` is in the range "[-61, 195]".
    k: int = ((value << 96) // 54916777467707473351141471128 + 2**95) >> 96
    value = value - (k * 54916777467707473351141471128)

    # Evaluate using a "(6, 7)"-term rational approximation. Since `p` is monic,
    # we will multiply by a scaling factor later.
    y: int = (
        ((value + 1346386616545796478920950773328) * value) >> 96
    ) + 57155421227552351082224309758442
    p: int = (
        ((((y + value) - 94201549194550492254356042504812) * y) >> 96)
        + 28719021644029726153956944680412240
    ) * value + (4385272521454847904659076985693276 << 96)

    # We leave `p` in the "2 ** 192" base so that we do not have to scale it up
    # again for the division.
    q: int = (
        ((value - 2855989394907223263936484059900) * value) >> 96
    ) + 50020603652535783019961831881945
    q = ((q * value) >> 96) - 533845033583426703283633433725380
    q = ((q * value) >> 96) + 3604857256930695427073651918091429
    q = ((q * value) >> 96) - 14423608567350463180887372962807573
    q = ((q * value) >> 96) + 26449188498355588339934803723976023

    # The polynomial `q` has no zeros in the range because all its roots are complex.
    # No scaling is required, as `p` is already "2 ** 96" too large. Also,
    # `r` is in the range "(0.09, 0.25) * 2**96" after the division.
    r: int = p // q

    # To finalise the calculation, we have to multiply `r` by:
    #   - the scale factor "s = ~6.031367120",
    #   - the factor "2 ** k" from the range reduction, and
    #   - the factor "1e18 / 2 ** 96" for the base conversion.
    # We do this all at once, with an intermediate result in "2**213" base,
    # so that the final right shift always gives a positive value.

    # Note that to circumvent Vyper's safecast feature for the potentially
    # negative parameter value `r`, we first convert `r` to `bytes32` and
    # subsequently to `uint256`. Remember that the EVM default behaviour is
    # to use two's complement representation to handle signed integers.
    return (r * 3822833074963236453042738258902158003155416615667) >> (195 - k)



"""Package to simulate Curve pool."""
__all__ = ["autosim", "bonding_curve", "order_book", "__version__", "__version_info__"]

from ._order_book import order_book
from .sim import autosim
from .tools import bonding_curve
from .version import __version__, __version_info__



from abc import ABC, abstractmethod

from curvesim.exceptions import ParameterSamplerError
from curvesim.logging import get_logger

logger = get_logger(__name__)


class ParameterSampler(ABC):
    """
    An iterator that yields pools with different parameter settings.
    """

    @abstractmethod
    def __iter__(self):
        """
        Yields
        -------
        pool : :class:`~curvesim.templates.SimPool`
            A pool object with the current variable parameters set.

        params : dict
            A dictionary of the pool parameters set on the current iteration.

        """
        raise NotImplementedError

    def set_pool_attributes(self, pool, attribute_dict):
        """
        Sets the pool attributes defined in attribute_dict.

        Supports setting attributes with :python:`setattr(pool, key, value)` or
        specialized setters defined in the 'setters' property:
        :python:`self.setters[key](pool, value)`

        For metapools, basepool parameters can be referenced by appending "_base" to
        an attribute's name.

        Parameters
        ----------
        pool : :class:`~curvesim.templates.SimPool`
            The pool object to be modified.

        attribute_dict : dict
            A dict mapping attribute names to values.
        """
        if attribute_dict is None:
            return

        self._validate_attributes(pool, attribute_dict)

        for attribute, value in attribute_dict.items():
            self._set_pool_attribute(pool, attribute, value)

    @property
    def setters(self):
        """
        A dict mapping attributes to setter functions.

        Used to set attributes that require more computation than simple setattr().
        Typically defined in pool-specific mixin. Defaults to empty dict.

        For metapools, basepool parameters can be referenced by appending "_base" to
        an attribute's name.

        Returns
        -------
        dict
        """
        return {}

    def _set_pool_attribute(self, pool, attr, value):
        """
        Sets a single pool attribute.

        Supports setting attributes with :python:`setattr(pool, attr, value)` or
        specialized setters defined in the 'setters' property:
        :python:`self.setters[attr](pool, value)`

        For metapools, basepool parameters can be referenced by appending "_base" to
        an attribute's name.

        Parameters
        ----------
        pool : :class:`~curvesim.templates.SimPool`
            The pool object to be modified.

        attr : str
            The attribute to be set.

        value :
            The value to be set for the attribute.
        """
        if attr in self.setters:
            self.setters[attr](pool, value)

        else:
            pool_attr = parse_pool_attribute(pool, attr)
            setattr(*pool_attr, value)

    def _validate_attributes(self, pool, attributes):
        """
        Raises error if attributes are not present in self.setters or pool attributes.

        Parameters
        ----------
        pool : :class:`~curvesim.templates.SimPool`
            The pool object to be modified.

        attributes : Iterable[str]
            Iterable of attribute names.

        Raises
        ------
        ParameterSamplerError

        """
        missing = []
        for attribute in attributes:
            pool_attr = parse_pool_attribute(pool, attribute)

            if attribute not in self.setters and not hasattr(*pool_attr):
                missing.append(attribute)

        if missing:
            pool_class = pool.__class__.__name__
            self_class = self.__class__.__name__

            raise ParameterSamplerError(
                f"Input parameters not found in '{self_class}.setters' "
                f"or '{pool_class}' attributes: {missing}"
            )


def parse_pool_attribute(pool, attribute):
    """
    Helper function to route "_base" attributes to basepool if necessary.
    """
    if attribute.endswith("_base"):

        if not hasattr(pool, "basepool"):
            pool_class = pool.__class__.__name__

            raise ParameterSamplerError(
                f"Could not set pool parameter '{attribute}'; "
                f"'{pool_class}' has no basepool."
            )

        return pool.basepool, attribute[:-5]

    return pool, attribute



"""
Specific metric classes for use in simulations.
"""


__all__ = [
    "ArbMetrics",
    "PoolBalance",
    "PoolVolume",
    "PoolValue",
    "PriceDepth",
    "Timestamp",
]

from copy import deepcopy

from altair import Axis, Scale
from numpy import array, exp, log, timedelta64
from pandas import DataFrame, concat

from curvesim.pool.sim_interface import (
    SimCurveCryptoPool,
    SimCurveMetaPool,
    SimCurvePool,
    SimCurveRaiPool,
)
from curvesim.utils import cache, get_pairs

from .base import Metric, PoolMetric, PoolPricingMetric, PricingMetric


class ArbMetrics(PricingMetric):
    """
    Computes metrics characterizing arbitrage trades: arbitrageur profits, pool fees,
    and post-trade price error between target and pool price.
    """

    @property
    def config(self):
        return {
            "functions": {
                "metrics": self.compute_arb_metrics,
                "summary": {
                    "arb_profit": "sum",
                    "pool_fees": "sum",
                    "price_error": "median",
                },
            },
            "plot": {
                "metrics": {
                    "arb_profit": {
                        "title": f"Daily Arbitrageur Profit (in {self.numeraire})",
                        "style": "time_series",
                        "resample": "sum",
                    },
                    "pool_fees": {
                        "title": f"Daily Pool Fees (in {self.numeraire})",
                        "style": "time_series",
                        "resample": "sum",
                    },
                    "price_error": {
                        "title": "Price Error",
                        "style": "histogram",
                        "encoding": {
                            "x": {
                                "title": "Price Error (binned)",
                                "shorthand": "price_error",
                                "scale": Scale(domain=[0, 0.05], clamp=True),
                            },
                        },
                    },
                },
                "summary": {
                    "arb_profit": {
                        "title": f"Total Arbitrageur Profit (in {self.numeraire})",
                        "style": "point_line",
                    },
                    "pool_fees": {
                        "title": f"Total Pool Fees (in {self.numeraire})",
                        "style": "point_line",
                    },
                    "price_error": {
                        "title": "Price Error (median)",
                        "style": "point_line",
                    },
                },
            },
        }

    def __init__(self, pool, **kwargs):
        super().__init__(pool.assets.symbols)

    def compute_arb_metrics(self, **kwargs):
        """Computes all metrics for each timestamp in an individual run."""
        price_sample = kwargs["price_sample"]
        trade_data = kwargs["trade_data"]

        prices = DataFrame(price_sample.prices.to_list(), index=price_sample.index)

        profits = self._compute_profits(prices, trade_data.trades)
        price_error = trade_data.price_errors.apply(
            lambda errors: sum(abs(e) for e in errors)
        )

        results = concat([profits, price_error], axis=1)
        results.columns = list(self.config["plot"]["metrics"])

        return results

    def _compute_profits(self, price_df, trade_df):
        """
        Computes arbitrageur profits and pool fees for a single row of data (i.e.,
        a single timestamp) in units of the chosen numeraire, `self.numeraire`.
        """
        numeraire = self.numeraire

        profit = []
        for price_row, trade_row in zip(price_df.iterrows(), trade_df):

            timestamp, prices = price_row
            arb_profit = 0
            pool_profit = 0

            for trade in trade_row:
                market_price = self.get_market_price(
                    trade.coin_in, trade.coin_out, prices
                )
                arb = trade.amount_out - trade.amount_in * market_price
                fee = trade.fee

                if trade.coin_out != numeraire:
                    price = self.get_market_price(trade.coin_out, numeraire, prices)
                    arb *= price
                    fee *= price

                arb_profit += arb
                pool_profit += fee

            profit.append(
                {
                    "timestamp": timestamp,
                    "arb_profit": arb_profit / 10**18,
                    "pool_profit": pool_profit / 10**18,
                }
            )

        return DataFrame(profit).set_index("timestamp")


class PoolVolume(PoolPricingMetric):
    """
    Records total trade volume for each timestamp.
    """

    @property
    @cache
    def pool_config(self):
        base = {
            "functions": {"summary": {"pool_volume": "sum"}},
            "plot": {
                "metrics": {
                    "pool_volume": {
                        "title": "Daily Volume",
                        "style": "time_series",
                        "resample": "sum",
                    },
                },
                "summary": {
                    "pool_volume": {
                        "title": "Total Volume",
                        "style": "point_line",
                    },
                },
            },
        }

        functions = {
            SimCurvePool: self.get_stableswap_pool_volume,
            SimCurveMetaPool: self.get_stableswap_metapool_volume,
            SimCurveRaiPool: self.get_stableswap_metapool_volume,
            SimCurveCryptoPool: self.get_cryptoswap_pool_volume,
        }

        units = {
            SimCurvePool: "(of Any Coin)",
            SimCurveMetaPool: "(of Any Coin)",
            SimCurveRaiPool: "(of Any Coin)",
            SimCurveCryptoPool: f"(in {self.numeraire})",
        }

        config = {}
        for pool in functions:
            cfg = deepcopy(base)
            cfg["functions"]["metrics"] = functions[pool]
            _units = units[pool]
            cfg["plot"]["metrics"]["pool_volume"]["title"] = "Daily Volume " + _units
            cfg["plot"]["summary"]["pool_volume"]["title"] = "Total Volume " + _units
            config[pool] = cfg

        return config

    def get_stableswap_pool_volume(self, **kwargs):
        """
        Records trade volume for stableswap non-meta-pools.
        """
        trade_data = kwargs["trade_data"]

        def per_timestamp_function(trade_data):
            trades = trade_data.trades
            return sum(trade.amount_in for trade in trades) / 10**18

        return self._get_volume(trade_data, per_timestamp_function)

    def get_stableswap_metapool_volume(self, **kwargs):
        """
        Records trade volume for stableswap meta-pools. Only includes trades involving
        the meta-asset (basepool-only trades are ignored).
        """
        trade_data = kwargs["trade_data"]

        meta_asset = self._pool.assets.symbols[0]

        def per_timestamp_function(trade_data):
            volume = 0
            for trade in trade_data.trades:
                if meta_asset in (trade.coin_in, trade.coin_out):
                    volume += trade.amount_in
            return volume / 10**18

        return self._get_volume(trade_data, per_timestamp_function)

    def get_cryptoswap_pool_volume(self, **kwargs):
        """
        Records trade volume for cryptoswap non-meta-pools.
        """
        trades = kwargs["trade_data"].trades
        prices = kwargs["price_sample"].prices

        trade_price_data = concat([trades, prices], axis=1)
        numeraire = self.numeraire

        def per_timestamp_function(trade_price_data):
            trades = trade_price_data.trades
            prices = trade_price_data.prices
            get_price = self.get_market_price

            volume = 0
            for trade in trades:
                volume += trade.amount_in * get_price(trade.coin_in, numeraire, prices)
            return volume / 10**18

        return self._get_volume(trade_price_data, per_timestamp_function)

    def _get_volume(self, trade_data, per_timestamp_function):
        volume = trade_data.apply(per_timestamp_function, axis=1)
        results = DataFrame(volume)
        results.columns = ["pool_volume"]
        return results


class PoolBalance(PoolMetric):
    """
    Computes the pool balance metric, which ranges from 0 (completely imbalanced) to 1
    (completely balanced).
    """

    @property
    @cache
    def pool_config(self):
        ss_config = {
            "functions": {
                "metrics": self.get_pool_balance,
                "summary": {"pool_balance": ["median", "min"]},
            },
            "plot": {
                "metrics": {
                    "pool_balance": {
                        "title": "Pool Balance/Imbalance",
                        "style": "time_series",
                        "resample": "median",
                        "encoding": {
                            "y": {
                                "title": "% Balanced (Daily Median)",
                                "axis": Axis(format="%"),
                            },
                        },
                    }
                },
                "summary": {
                    "pool_balance": {
                        "title": "Pool Balance/Imbalance",
                        "style": "point_line",
                        "encoding": {
                            "y": {"title": "% Balanced", "axis": Axis(format="%")},
                        },
                    }
                },
            },
        }

        return dict.fromkeys(
            [SimCurveMetaPool, SimCurvePool, SimCurveRaiPool, SimCurveCryptoPool],
            ss_config,
        )

    def get_pool_balance(self, **kwargs):
        """
        Computes pool balance metrics for each timestamp in an individual run.
        Used for any Curve pool.
        """
        pool_state = kwargs["pool_state"]
        balance = pool_state.apply(self._compute_stableswap_balance, axis=1)
        return DataFrame(balance, columns=["pool_balance"])

    def _compute_stableswap_balance(self, pool_state_row):
        """
        Computes balance metric for a single row of data (i.e., a single timestamp).
        Used for any Curve pool.
        """
        self.set_pool_state(pool_state_row)
        pool = self._pool

        xp = array(pool._xp())  # pylint: disable=protected-access
        n = pool.n
        bal = 1 - sum(abs(xp / sum(xp) - 1 / n)) / (2 * (n - 1) / n)

        return bal


class PoolValue(PoolPricingMetric):
    """
    Computes pool's value over time in virtual units and the chosen
    numeraire, `self.numeraire`. Each are summarized as annualized returns.
    """

    @property
    @cache
    def pool_config(self):
        plot = {
            "metrics": {
                "pool_value_virtual": {
                    "title": "Pool Value (Virtual)",
                    "style": "time_series",
                    "resample": "last",
                },
                "pool_value": {
                    "title": f"Pool Value (in {self.numeraire})",
                    "style": "time_series",
                    "resample": "last",
                },
            },
            "summary": {
                "pool_value_virtual": {
                    "title": "Annualized Returns (Virtual)",
                    "style": "point_line",
                    "encoding": {"y": {"axis": Axis(format="%")}},
                },
                "pool_value": {
                    "title": f"Annualized Returns (in {self.numeraire})",
                    "style": "point_line",
                    "encoding": {"y": {"axis": Axis(format="%")}},
                },
            },
        }

        summary_fns = {
            "pool_value_virtual": {
                "annualized_returns": self.compute_annualized_returns
            },
            "pool_value": {"annualized_returns": self.compute_annualized_returns},
        }

        base = {
            "functions": {"summary": summary_fns},
            "plot": plot,
        }

        functions = {
            SimCurvePool: self.get_stableswap_pool_value,
            SimCurveMetaPool: self.get_stableswap_metapool_value,
            SimCurveRaiPool: self.get_stableswap_metapool_value,
            SimCurveCryptoPool: self.get_cryptoswap_pool_value,
        }

        config = {}
        for pool, fn in functions.items():
            config[pool] = deepcopy(base)
            config[pool]["functions"]["metrics"] = fn

        return config

    def get_stableswap_pool_value(self, **kwargs):
        """
        Computes all metrics for each timestamp in an individual run.
        Used for non-meta stableswap pools.
        """

        return self._get_pool_value(
            kwargs["pool_state"],
            kwargs["price_sample"],
            self._get_stableswap_virtual_value,
        )

    def get_stableswap_metapool_value(self, **kwargs):
        """
        Computes all metrics for each timestamp in an individual run.
        Used for stableswap metapools.
        """

        return self._get_metapool_value(
            kwargs["pool_state"],
            kwargs["price_sample"],
            self._get_stableswap_virtual_value,
        )

    def get_cryptoswap_pool_value(self, **kwargs):
        """
        Computes all metrics for each timestamp in an individual run.
        Used for non-meta cryptoswap pools.
        """

        return self._get_pool_value(
            kwargs["pool_state"],
            kwargs["price_sample"],
            self._get_cryptoswap_virtual_value,
        )

    def _get_pool_value(self, pool_state, price_sample, virtual_price_fn):
        """
        Computes all metrics for each timestamp in an individual run.
        Used for non-meta pools.
        """
        reserves = DataFrame(
            pool_state.balances.to_list(),
            index=pool_state.index,
            columns=self._pool.coin_names,
        )

        prices = DataFrame(price_sample.prices.to_list(), index=price_sample.index)

        pool_value = self._get_value_from_prices(reserves / 10**18, prices)
        pool_value_virtual = pool_state.apply(virtual_price_fn, axis=1)

        results = concat([pool_value_virtual, pool_value], axis=1)
        results.columns = list(self.config["plot"]["metrics"])
        return results.astype("float64")

    def _get_metapool_value(self, pool_state, price_sample, virtual_price_fn):
        """
        Computes all metrics for each timestamp in an individual run.
        Used for stableswap metapools.
        """
        max_coin = self._pool.max_coin
        pool = self._pool

        meta_reserves = DataFrame(
            pool_state.balances.to_list(),
            index=pool_state.index,
            columns=pool.coin_names,
        )

        base_reserves = DataFrame(
            pool_state.balances_base.to_list(),
            index=pool_state.index,
            columns=pool.basepool.coin_names,
        )

        prices = DataFrame(price_sample.prices.to_list(), index=price_sample.index)

        LP_token_proportion = meta_reserves.iloc[:, max_coin] / pool_state.tokens_base
        base_reserves = base_reserves.mul(LP_token_proportion, axis=0)
        reserves = concat([meta_reserves.iloc[:, :max_coin], base_reserves], axis=1)

        pool_value = self._get_value_from_prices(reserves / 10**18, prices)
        pool_value_virtual = pool_state.apply(virtual_price_fn, axis=1)

        results = concat([pool_value_virtual, pool_value], axis=1)
        results.columns = list(self.config["plot"]["metrics"])
        return results.astype("float64")

    def _get_value_from_prices(self, reserves, prices):
        """
        Computes pool value in the chosen numeraire, `self.numeraire`.
        Can be used for any pool type.
        """
        get_price = self.get_market_price
        numeraire = self.numeraire

        value = 0
        for coin_name in reserves.columns:
            value += reserves[coin_name] * get_price(coin_name, numeraire, prices)
        return value

    def _get_stableswap_virtual_value(self, pool_state_row):
        """
        Computes virtual pool value for a single row of data (i.e., a single timestamp).
        Used for any stableswap pool.
        """
        self.set_pool_state(pool_state_row)
        return self._pool.D() / 10**18

    def _get_cryptoswap_virtual_value(self, pool_state_row):
        """
        Computes virtual pool value for a single row of data (i.e., a single timestamp).
        Used for any cryptoswap pool.
        """
        self.set_pool_state(pool_state_row)
        # pylint: disable-next=protected-access
        return self._pool._get_xcp(self._pool.D) / 10**18

    def compute_annualized_returns(self, data):
        """Computes annualized returns from a series of pool values."""
        year_multipliers = timedelta64(1, "Y") / data.index.to_series().diff()
        log_returns = log(data).diff()  # pylint: disable=no-member

        return exp((log_returns * year_multipliers).mean()) - 1


class PriceDepth(PoolMetric):
    """
    Computes metrics indicating a pool's price (liquidity) depth. Generally, uses
    liquidity density, % change in reserves per % change in price.
    """

    __slots__ = ["_factor"]

    @property
    @cache
    def pool_config(self):
        ss_config = {
            "functions": {
                "metrics": self.get_curve_LD,
                "summary": {"liquidity_density": ["median", "min"]},
            },
            "plot": {
                "metrics": {
                    "liquidity_density": {
                        "title": "Liquidity Density (Daily Median)",
                        "style": "time_series",
                        "resample": "median",
                        "encoding": {
                            "y": {"title": "Liquidity Density (Daily Median)"}
                        },
                    }
                },
                "summary": {
                    "liquidity_density": {
                        "title": "Liquidity Density",
                        "style": "point_line",
                    }
                },
            },
        }

        return dict.fromkeys(
            [SimCurveMetaPool, SimCurvePool, SimCurveRaiPool, SimCurveCryptoPool],
            ss_config,
        )

    def __init__(self, pool, factor=10**8, **kwargs):
        self._factor = factor
        super().__init__(pool, **kwargs)

    def get_curve_LD(self, **kwargs):
        """
        Computes liquidity density for each timestamp in an individual run.
        Used for all Curve pools.
        """
        pool_state = kwargs["pool_state"]

        coin_pairs = get_pairs(
            self._pool.coin_names
        )  # for metapool, uses only meta assets
        LD = pool_state.apply(self._get_curve_LD_by_row, axis=1, coin_pairs=coin_pairs)
        return DataFrame(LD, columns=["liquidity_density"])

    def _get_curve_LD_by_row(self, pool_state_row, coin_pairs):
        """
        Computes liquidity density for a single row of data (i.e., a single timestamp).
        Used for all Curve pools.
        """
        self.set_pool_state(pool_state_row)

        LD = []
        for pair in coin_pairs:
            ld = self._compute_liquidity_density(*pair)
            LD.append(ld)
        return sum(LD) / len(LD)

    def _compute_liquidity_density(self, coin_in, coin_out):
        """
        Computes liquidity density for a single pair of coins.
        """
        factor = self._factor
        pool = self._pool
        post_trade_price = self._post_trade_price

        price_pre = pool.price(coin_in, coin_out, use_fee=False)
        price_post = post_trade_price(pool, coin_in, coin_out, factor)
        LD1 = price_pre / ((price_pre - price_post) * factor)

        price_pre = pool.price(coin_out, coin_in, use_fee=False)
        # pylint: disable-next=arguments-out-of-order
        price_post = post_trade_price(pool, coin_out, coin_in, factor)
        LD2 = price_pre / ((price_pre - price_post) * factor)

        return (LD1 + LD2) / 2

    @staticmethod
    def _post_trade_price(pool, coin_in, coin_out, factor, use_fee=False):
        """
        Computes price after executing a trade of size coin_in balances / factor.
        """

        size = pool.asset_balances[coin_in] // factor

        with pool.use_snapshot_context():
            pool.trade(coin_in, coin_out, size)
            price = pool.price(coin_in, coin_out, use_fee=use_fee)

        return price


class Timestamp(Metric):
    """Simple pass-through metric to record timestamps."""

    @property
    def config(self):
        return {"functions": {"metrics": self._get_timestamp}}

    def _get_timestamp(self, **kwargs):
        price_sample = kwargs["price_sample"]
        return DataFrame(price_sample.timestamp)



"""
Contains the bonding_curve function, which computes a pool's bonding
curve and current reserves for each pair of coins and optionally
plots the curves using Matplotlib.
"""
from itertools import combinations

import matplotlib.pyplot as plt
from numpy import linspace

from curvesim.pool import CurveMetaPool

D_UNIT = 10**18


def bonding_curve(pool, *, truncate=0.0005, resolution=1000, plot=False):
    """
    Computes and optionally plots a pool's bonding curve and current reserves.

    Parameters
    ----------
    pool : CurvePool or CurveMetaPool
        The pool object for which the bonding curve is computed.

    truncate : float, optional (default=0.0005)
        Determines where to truncate the bonding curve. The truncation point is given
        by D*truncate, where D is the total supply of tokens in the pool.

    resolution : int, optional (default=1000)
        The number of points to compute along the bonding curve.

    plot : bool, optional (default=False)
        Plots the bonding curves using Matplotlib.

    Returns
    -------
    pair_to_curve : dict
        Dictionary with coin index pairs as keys and lists of corresponding reserves
        as values. Each list of reserves is a list of pairs, where each pair consists
        of the reserves for the first and second coin of the corresponding pair.

    Example
    --------
    >>> import curvesim
    >>> pool_address = "0xbEbc44782C7dB0a1A60Cb6fe97d0b483032FF1C7"
    >>> pool = curvesim.pool.get(pool_address)
    >>> pair_to_curve = curvesim.bonding_curve(pool, plot=True)
    """

    if isinstance(pool, CurveMetaPool):
        combos = [(0, 1)]
    else:
        combos = combinations(range(pool.n), 2)

    D = pool.D()
    xp = pool._xp()  # pylint: disable=protected-access

    pair_to_curve = {}
    for (i, j) in combos:
        truncated_D = int(D * truncate)
        x_max = pool.get_y(j, i, truncated_D, xp)
        xs = linspace(truncated_D, x_max, resolution).round()

        curve = []
        for x in xs:
            y = pool.get_y(i, j, int(x), xp)
            curve.append((x, y))
        curve = [(x / D_UNIT, y / D_UNIT) for x, y in curve]
        pair_to_curve[(i, j)] = curve

    if plot:
        labels = pool.coin_names
        if not labels:
            labels = [f"Coin {str(label)}" for label in range(pool.n)]

        _plot_bonding_curve(pair_to_curve, labels, xp)

    return pair_to_curve


def _plot_bonding_curve(pair_to_curve, labels, xp):
    n = len(pair_to_curve)
    _, axs = plt.subplots(1, n, constrained_layout=True)
    if n == 1:
        axs = [axs]

    for pair, ax in zip(pair_to_curve, axs):
        curve = pair_to_curve[pair]
        xs, ys = zip(*curve)
        ax.plot(xs, ys, color="black")

        i, j = pair
        ax.scatter(xp[i] / D_UNIT, xp[j] / D_UNIT, s=40, color="black")
        ax.set_xlabel(labels[i])
        ax.set_ylabel(labels[j])

    plt.show()



from abc import ABC, abstractmethod

from curvesim.logging import get_logger

logger = get_logger(__name__)


class Strategy(ABC):
    """
    A Strategy defines the trading approach used during each step of a simulation.
    It executes the trades using an injected `Trader` class and then logs the
    changes using the injected `StateLog` class.

    Class Attributes
    ----------------
    trader_class : :class:`~curvesim.pipelines.templates.Trader`
        Class for creating trader instances.
    state_log_class : :class:`~curvesim.metrics.StateLog`
        Class for creating state logger instances.

    Attributes
    ----------
    metrics : List[Metric]
        A list of metrics used to evaluate the performance of the strategy.
    """

    # These classes should be injected in child classes
    # to create the desired behavior.
    trader_class = None
    state_log_class = None

    def __init__(self, metrics):
        """
        Parameters
        ----------
        metrics : List[Metric]
            A list of metrics used to evaluate the performance of the strategy.
        """
        self.metrics = metrics

    def __call__(self, pool, parameters, price_sampler):
        """
        Computes and executes trades at each timestep.

        Parameters
        ----------
        pool : :class:`~curvesim.pipelines.templates.SimPool`
            The pool to be traded against.

        parameters : dict
            Current pool parameters from the param_sampler (only used for
            logging/display).

        price_sampler : iterable
            Iterable that for each timestep returns market data used by
            the trader.


        Returns
        -------
        metrics : tuple of lists

        """
        # pylint: disable=not-callable
        trader = self.trader_class(pool)
        state_log = self.state_log_class(pool, self.metrics)

        logger.info("[%s] Simulating with %s", pool.symbol, parameters)

        pool.prepare_for_run(price_sampler.prices)

        for sample in price_sampler:
            pool.prepare_for_trades(sample.timestamp)
            trader_args = self._get_trader_inputs(sample)
            trade_data = trader.process_time_sample(*trader_args)
            state_log.update(price_sample=sample, trade_data=trade_data)

        return state_log.compute_metrics()

    @abstractmethod
    def _get_trader_inputs(self, sample):
        """
        Process the price sample into appropriate inputs for the
        trader instance.
        """
        raise NotImplementedError



"""
This module provides functionality for creating interactive Altair charts from
simulation results.

It contains utility functions for creating data dictionaries, pages, subplots,
and layered or single charts. It uses selectors to control chart properties and
apply tooltips.
"""
from altair import concat, data_transformers, layer, value, vconcat

from ..make_chart import make_chart
from ..selectors import if_selected, make_selector
from .preprocessing import preprocess_data
from .result_selectors import make_result_selectors
from .tooltip import make_tooltip

data_transformers.disable_max_rows()


def make_page_from_results(results, data_key, axes, downsample=False):
    """
    Creates an interactive Altair chart from simulation results.

    Parameters
    ----------
    results : Results
        The simulation results.
    data_key : str
        The key to retrieve data from the results.
    axes : dict
        The axes configuration.
    downsample : bool, optional
        Whether to downsample the data. Defaults to False.

    Returns
    -------
    altair.vconcat
        The created chart page.
    """
    config = results.plot_config[data_key]
    factor_dict = results.factors
    factor_keys = list(factor_dict.keys())

    data_dict = make_data_dict(results, data_key, config, factor_keys, downsample)
    selectors = make_result_selectors(factor_dict, axes["dynamic"])
    return make_page(data_dict, config, factor_keys, axes["metric"], selectors)


def make_data_dict(results, data_key, config, factor_keys, downsample):
    """
    Prepares the data from the results for chart creation.

    Parameters
    ----------
    results : Results
        The simulation results.
    data_key : str
        The key to retrieve data from the results.
    config : dict
        The plot configuration.
    factor_keys : list
        The keys for the factors.
    downsample : bool
        Whether to downsample the data.

    Returns
    -------
    dict
        The data dictionary.
    """
    columns = factor_keys + list(config.keys())
    data = getattr(results, data_key)(columns=columns)
    return preprocess_data(data, config, factor_keys, downsample)


def make_page(data_dict, config, factors, metric_axis, selectors):
    """
    Creates an interactive Altair chart page with multiple subplots.

    Parameters
    ----------
    data_dict : dict
        The data dictionary.
    config : dict
        The plot configuration.
    factors : list
        The list of factors.
    metric_axis : str
        The axis for the metrics.
    selectors : dict
        The selectors configuration.

    Returns
    -------
    altair.vconcat
        The created chart page.
    """
    kwargs = selectors["kwargs"]
    page = concat(data=data_dict["main"], columns=2)

    for metric_key, cfg in config.items():
        metrics, kwargs["data"] = get_metric_data(metric_key, data_dict)
        subplot = make_subplot(cfg, metrics, factors, metric_axis, kwargs)
        page |= subplot

    if factors:
        page = vconcat(selectors["charts"], page)

    return page.resolve_scale(color="independent")


def get_metric_data(metric_key, data_dict):
    """
    Retrieves the relevant data for a specific metric.

    Parameters
    ----------
    metric_key : str
        The key for the metric.
    data_dict : dict
        The data dictionary.

    Returns
    -------
    str or list of str, pandas.DataFrame
        The metrics and the data for the metrics.
    """
    if metric_key in data_dict:
        data = data_dict[metric_key]
    else:
        data = data_dict["main"]

    submetrics = [c for c in data if c.startswith(metric_key + " ")]
    n_submetrics = len(submetrics)

    if n_submetrics == 0:
        return metric_key, data

    if n_submetrics == 1:
        return submetrics[0], data

    return submetrics, data


def make_subplot(config, metrics, factors, metric_axis, kwargs):
    """
    Creates a subplot for a specific metric.

    Parameters
    ----------
    config : dict
        The plot configuration.
    metrics : str or list of str
        The metric or metrics.
    factors : list
        The list of factors.
    metric_axis : str
        The axis for the metrics.
    kwargs : dict
        Additional keyword arguments.

    Returns
    -------
    altair.vconcat
        The created subplot.
    """
    if isinstance(metrics, list):
        make = make_layered_chart
    else:
        make = make_single_chart

    return make(config, metrics, metric_axis, factors, **kwargs)


def make_single_chart(config, metric, metric_axis, factors, **kwargs):
    """
    Creates a single chart for a specific metric.

    Parameters
    ----------
    config : dict
        The plot configuration.
    metric : str
        The metric.
    metric_axis : str
        The axis for the metrics.
    factors : list
        The list of factors.
    **kwargs
        Additional keyword arguments.

    Returns
    -------
    altair.Chart
        The created chart.
    """
    kwargs[metric_axis] = metric
    chart = make_chart(config, **kwargs)
    tooltip = make_tooltip(chart["encoding"], metric_axis, factors)
    return chart.encode(tooltip=tooltip)


def make_layered_chart(config, metrics, metric_axis, factors, **kwargs):
    """
    Creates a layered chart for multiple submetrics of a metric.

    Parameters
    ----------
    config : dict
        The plot configuration.
    metrics : list of str
        The metrics.
    metric_axis : str
        The axis for the metrics.
    factors : list
        The list of factors.
    **kwargs
        Additional keyword arguments.

    Returns
    -------
    altair.vconcat
        The created chart.
    """
    labels = [metric.split(" ")[1].capitalize() for metric in metrics]
    sel_chart, selector = make_selector(
        "submetric", metrics, labels=labels, toggle="true", sel_idx=0
    )

    chart = layer()
    for submetric, label in zip(metrics, labels):
        kwargs[metric_axis] = submetric
        _layer = make_chart(config, **kwargs)
        tooltip = make_tooltip(_layer["encoding"], metric_axis, factors, prefix=label)
        opacity = if_selected(submetric, selector, "submetric", value(1), value(0))
        chart += _layer.encode(tooltip=tooltip, opacity=opacity)
    return vconcat(chart, sel_chart, spacing=2)



from abc import abstractmethod
from copy import deepcopy
from itertools import product

from curvesim.exceptions import ParameterSamplerError
from curvesim.pool.sim_interface import (
    SimCurveCryptoPool,
    SimCurveMetaPool,
    SimCurvePool,
    SimCurveRaiPool,
)
from curvesim.templates import ParameterSampler

from .pool_mixins import CurveCryptoPoolMixin, CurveMetaPoolMixin, CurvePoolMixin


class ParameterizedPoolIterator(ParameterSampler):
    """
    Iterates over pools with all possible combinations of the input parameters.
    """

    # pylint: disable-next=unused-argument
    def __new__(cls, pool, variable_params=None, fixed_params=None, pool_map=None):
        """
        Returns a pool-specific ParameterizedPoolIterator subclass.

        Parameters
        ----------
        pool_map : dict, optional
            A mapping between pool types and subclasses. Overrides default mapping.

        Returns
        -------
            :class:`.ParameterizedPoolIterator` subclass

        """
        pool_map = pool_map or DEFAULT_POOL_MAP

        if cls is not ParameterizedPoolIterator:
            return super().__new__(cls)

        try:
            pool_type = type(pool)
            subclass = pool_map[pool_type]

        except KeyError as e:
            pool_type_name = pool_type.__name__
            raise ParameterSamplerError(
                f"No subclass for pool type `{pool_type_name}` found in "
                "ParameterizedPoolIterator pool map."
            ) from e

        return super().__new__(subclass)

    # pylint: disable-next=unused-argument
    def __init__(self, pool, variable_params=None, fixed_params=None, pool_map=None):
        """
        Parameters
        ----------
        pool : :class:`~curvesim.templates.SimPool`
            The "template" pool that will have its parameters modified.

        variable_params: dict, optional
            Pool parameters to vary across simulations.

            Keys are parameter names and values are iterables of values. For metapools,
            basepool parameters can be referenced by appending "_base" to an attribute
            name.

            Example
            --------
            .. code-block ::

                {"A": [100, 1000], "fee_base": [10**6, 4*10**6]}

        fixed_params : dict, optional
            Pool parameters set before all simulations.

        pool_map : dict, optional
            See __new__ method.

        """
        self._validate_pool_type(pool)
        self.pool_template = deepcopy(pool)
        self.set_pool_attributes(self.pool_template, fixed_params)
        parameter_sequence = self.make_parameter_sequence(variable_params)
        self.parameter_sequence = parameter_sequence or [fixed_params]

    def __iter__(self):
        """
        Yields
        -------
        pool : :class:`~curvesim.templates.SimPool`
            A pool object with the current variable parameters set.

        params : dict
            A dictionary of the pool parameters set on this iteration.
        """
        for params in self.parameter_sequence:
            pool = deepcopy(self.pool_template)
            self.set_pool_attributes(pool, params)
            yield pool, params

    def make_parameter_sequence(self, variable_params):
        """
        Returns a list of dicts for each possible combination of the input parameters.

        Parameters
        ----------
        variable_params: dict
            Pool parameters to vary across simulations.

            Keys: pool parameters, Values: iterable of values

        Returns
        -------
        List(dict)
            A list of dicts defining the parameters for each iteration.
        """
        if not variable_params:
            return []

        keys, values = zip(*variable_params.items())
        self._validate_attributes(self.pool_template, keys)

        sequence = []
        for vals in product(*values):
            params = dict(zip(keys, vals))
            sequence.append(params)

        return sequence

    def _validate_pool_type(self, pool):
        """Validates that the input pool is an instance of self._pool_type."""
        if not isinstance(pool, self._pool_type):
            input_class = pool.__class__.__name__
            expected_class = self._pool_type.__name__
            self_class = self.__class__.__name__

            raise ParameterSamplerError(
                f"Parameter sampler '{self_class}' only supports pool type "
                f"'{expected_class}'; recieved '{input_class}'."
            )

    @property
    @abstractmethod
    def _pool_type(self):
        """The expected pool type for a ParameterizedPoolIterator subclass."""
        raise NotImplementedError


class ParameterizedCurvePoolIterator(CurvePoolMixin, ParameterizedPoolIterator):
    """
    :class:`ParameterizedPoolIterator` parameter sampler specialized
    for Curve pools.
    """


class ParameterizedCurveMetaPoolIterator(CurveMetaPoolMixin, ParameterizedPoolIterator):
    """
    :class:`ParameterizedPoolIterator` parameter sampler specialized
    for Curve meta-pools.
    """


class ParameterizedCurveCryptoPoolIterator(
    CurveCryptoPoolMixin, ParameterizedPoolIterator
):
    """
    :class:`ParameterizedPoolIterator` parameter sampler specialized
    for Curve crypto pools.
    """


DEFAULT_POOL_MAP = {
    SimCurvePool: ParameterizedCurvePoolIterator,
    SimCurveRaiPool: ParameterizedCurveMetaPoolIterator,
    SimCurveMetaPool: ParameterizedCurveMetaPoolIterator,
    SimCurveCryptoPool: ParameterizedCurveCryptoPoolIterator,
}



from .metapool import CurveMetaPool


class CurveRaiPool(CurveMetaPool):
    """
    Rebasing stableswap metapool implementation in Python. Used for RAI3CRV pool.

    """

    def __init__(self, redemption_price, *args, **kwargs):
        """
        Parameters
        ----------
        redemption_price : int
            redemption price for the pool; functionally equivalent to `rate_multiplier`
            for a factory metapool
        A : int
            Amplification coefficient; this is :math:`A n^{n-1}` in the whitepaper.
        D : int or list of int
            coin balances or virtual total balance
        n: int
            number of coins
        tokens : int
            LP token supply
        fee : int, optional
            fee with 10**10 precision (default = .004%)
        fee_mul :
            fee multiplier for dynamic fee pools
        admin_fee : int, optional
            percentage of `fee` with 10**10 precision (default = 50%)
        """
        super().__init__(*args, rate_multiplier=redemption_price, **kwargs)

    def dydx(self, i, j, use_fee=False):
        _dydx = super().dydx(i, j, use_fee=use_fee)

        if i >= self.max_coin and j == 0:
            base_i = i - self.max_coin
            _dydx = _dydx * self.basepool.rates[base_i] / self.rate_multiplier

        return _dydx

    def _dydx(self, i, j, xp, use_fee=False):
        dydx = super()._dydx(i, j, xp, use_fee=use_fee)
        rates = self.rates
        return dydx * rates[i] / rates[j]



""" Custom logging config and useful aliases for logging """

__all__ = ["get_logger", "multiprocessing_logging_queue"]

import datetime
import logging
import logging.config
import multiprocessing as mp
import os
from contextlib import contextmanager
from logging.handlers import QueueHandler, QueueListener

# -- convenient parameters to adjust for debugging -- #
DEFAULT_LEVEL = "info"
USE_LOG_FILE = False
# --------------------------------------------------- #

LEVELS = {
    "critical": logging.CRITICAL,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "info": logging.INFO,
    "debug": logging.DEBUG,
}


BASE_DIR = os.getcwd()
LOG_DIR = os.path.join(BASE_DIR, "logs")
if USE_LOG_FILE:
    os.makedirs(LOG_DIR, exist_ok=True)

__dt_string = datetime.datetime.now().strftime("%Y%m%d")
LOG_FILEPATH = os.path.join(LOG_DIR, __dt_string + ".log")

LOGGING_FORMAT = "[%(levelname)s][%(asctime)s][%(name)s]: %(message)s"
MULTIPROCESS_LOGGING_FORMAT = (
    "[%(levelname)s][%(asctime)s][%(name)s]-%(process)d: %(message)s"
)


# FIXME: need a function to update the config after module initialization
HANDLERS = ["console", "file"] if USE_LOG_FILE else ["console"]

CUSTOM_LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {"format": MULTIPROCESS_LOGGING_FORMAT, "datefmt": "%H:%M:%S"},
    },
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "standard",
        },
        "file": {
            "level": "DEBUG",
            "class": "logging.handlers.RotatingFileHandler",
            "maxBytes": 10 * 1024 * 1024,
            "mode": "a",
            "backupCount": 10,
            "filename": LOG_FILEPATH,
            "formatter": "standard",
            "delay": True,
        },
    },
    "loggers": {
        "": {
            "handlers": HANDLERS,
            "level": "DEBUG",
            "propagate": False,
        },
    },
}

# 3rd party loggers that we want to largely ignore
silenced_loggers = ["matplotlib", "asyncio", "rlp", "parso", "web3"]
configured_loggers = CUSTOM_LOGGING_CONFIG["loggers"]
for name in silenced_loggers:
    configured_loggers[name] = {
        "handlers": HANDLERS,
        "level": "INFO",
        "propagate": False,
    }

logging.config.dictConfig(CUSTOM_LOGGING_CONFIG)


def get_logger(logger_name, level=DEFAULT_LEVEL):
    """
    Ensures logging config is loaded and allows us
    to make various customizations.
    """
    logger = logging.getLogger(logger_name)
    if isinstance(level, str):
        level = LEVELS[level.strip().lower()]
    logger.setLevel(level)

    return logger


@contextmanager
def multiprocessing_logging_queue():
    """
    Context manager for a logging queue that can be shared
    across multiple processes.
    """
    with mp.Manager() as manager:
        try:
            logging_queue = manager.Queue()
            root_logger = get_logger("")
            listener = QueueListener(logging_queue, *root_logger.handlers)
            listener.start()
            yield logging_queue
        finally:
            listener.stop()


def configure_multiprocess_logging(logging_queue):
    """Configure root logger in process to enqueue logs."""
    root_logger = get_logger("")
    root_logger.handlers.clear()
    root_logger.addHandler(QueueHandler(logging_queue))



from typing import List

from gmpy2 import mpz

from curvesim.exceptions import CalculationError, CurvesimValueError
from curvesim.logging import get_logger

logger = get_logger(__name__)


MIN_GAMMA = 10**10
MAX_GAMMA = 2 * 10**16


PRECISION = 10**18  # The precision to convert to
A_MULTIPLIER = 10000

N_COINS = 2
MIN_A = N_COINS**N_COINS * A_MULTIPLIER // 10
MAX_A = N_COINS**N_COINS * A_MULTIPLIER * 100000


def geometric_mean(unsorted_x: List[int], sort: bool) -> int:
    """
    (x[0] * x[1] * ...) ** (1/N)
    """
    n_coins: int = len(unsorted_x)
    x: List[int] = unsorted_x
    if sort:
        x = sorted(unsorted_x, reverse=True)

    D: int = mpz(x[0])
    diff: int = 0
    for _ in range(255):
        D_prev: int = D
        D = (D + x[0] * x[1] // D) // n_coins
        diff = abs(D_prev - D)
        if diff <= 1 or diff * 10**18 < D:
            return int(D)
    raise CalculationError("Did not converge")


def _sqrt_int(x: int) -> int:
    """
    Originating from: https://github.com/vyperlang/vyper/issues/1266
    """
    if x == 0:
        return 0

    x = mpz(x)
    z: int = (x + 10**18) // 2
    y: int = x

    for _ in range(256):
        if z == y:
            return int(y)
        y = z
        z = (x * 10**18 // z + z) // 2

    raise CalculationError("Did not converge")


def lp_price(virtual_price, price_oracle) -> int:
    """
    Returns an LP token price approximating behavior as a constant-product AMM.
    """
    price_oracle = price_oracle[0]
    return 2 * virtual_price * _sqrt_int(price_oracle) // 10**18


def newton_y(  # noqa: complexity: 11
    ANN: int,
    gamma: int,
    x: List[int],
    D: int,
    i: int,
) -> int:
    """
    Calculating x[i] given other balances x[0..n_coins-1] and invariant D
    ANN = A * N**N
    """
    n_coins: int = len(x)

    # Safety checks
    if not MIN_A <= ANN <= MAX_A:
        raise CurvesimValueError("Unsafe value for A")
    if not MIN_GAMMA <= gamma <= MAX_GAMMA:
        raise CurvesimValueError("Unsafe value for gamma")
    assert 10**17 <= D <= 10**15 * 10**18
    for k in range(n_coins):
        if k != i:
            frac: int = x[k] * 10**18 // D
            assert 10**16 <= frac <= 10**20

    y: int = D // n_coins
    K0_i: int = 10**18
    S_i: int = 0

    x_sorted: List[int] = x.copy()
    x_sorted[i] = 0
    x_sorted = sorted(x_sorted, reverse=True)  # From high to low

    convergence_limit: int = max(max(x_sorted[0] // 10**14, D // 10**14), 100)

    # Formula for 2 coins
    S_i: int = x[1 - i]
    y: int = D**2 // (S_i * n_coins**2)
    K0_i: int = (10**18 * n_coins) * S_i // D

    y = mpz(y)
    K0_i = mpz(K0_i)
    D = mpz(D)
    for _ in range(255):
        y_prev: int = y

        K0: int = K0_i * y * n_coins // D
        S: int = S_i + y

        _g1k0: int = abs(gamma + 10**18 - K0) + 1

        # D / (A * N**N) * _g1k0**2 / gamma**2
        mul1: int = 10**18 * D // gamma * _g1k0 // gamma * _g1k0 * A_MULTIPLIER // ANN

        # 2*K0 / _g1k0
        mul2: int = 10**18 + (2 * 10**18) * K0 // _g1k0

        yfprime: int = 10**18 * y + S * mul2 + mul1
        _dyfprime: int = D * mul2
        if yfprime < _dyfprime:
            y = y_prev // 2
            continue

        yfprime -= _dyfprime
        fprime: int = yfprime // y

        # y -= f / f_prime;  y = (y * fprime - f) / fprime
        # y = (yfprime + 10**18 * D - 10**18 * S)
        #   / fprime + mul1 / fprime * (10**18 - K0) / K0
        y_minus: int = mul1 // fprime
        y_plus: int = (yfprime + 10**18 * D) // fprime + y_minus * 10**18 // K0
        y_minus += 10**18 * S // fprime

        if y_plus < y_minus:
            y = y_prev // 2
        else:
            y = y_plus - y_minus

        diff: int = abs(y - y_prev)
        if diff < max(convergence_limit, y // 10**14):
            frac: int = y * 10**18 // D
            assert 10**16 <= frac <= 10**20  # dev: unsafe value for y
            return int(y)

    raise CalculationError("Did not converge")


# pylint: disable=too-many-locals,too-many-branches
def newton_D(  # noqa: complexity: 13
    ANN: int,
    gamma: int,
    x_unsorted: List[int],
) -> List[int]:
    """
    Finding the `D` invariant using Newton's method.

    ANN is A * N**N from the whitepaper multiplied by the
    factor A_MULTIPLIER.
    """
    n_coins: int = len(x_unsorted)

    # Safety checks
    if not MIN_A <= ANN <= MAX_A:
        raise CurvesimValueError("Unsafe value for A")
    if not MIN_GAMMA <= gamma <= MAX_GAMMA:
        raise CurvesimValueError("Unsafe value for gamma")

    x: List[int] = sorted(x_unsorted, reverse=True)

    assert 10**9 <= x[0] <= 10**15 * 10**18
    for i in range(1, n_coins):
        frac: int = x[i] * 10**18 // x[0]
        assert frac >= 10**11

    D: int = n_coins * geometric_mean(x, False)
    S: int = sum(x)

    D = mpz(D)
    S = mpz(S)
    for _ in range(255):
        D_prev: int = D

        # formula for 2 coins
        K0: int = (10**18 * n_coins**2) * x[0] // D * x[1] // D

        _g1k0: int = abs(gamma + 10**18 - K0) + 1

        # D / (A * N**N) * _g1k0**2 / gamma**2
        mul1: int = 10**18 * D // gamma * _g1k0 // gamma * _g1k0 * A_MULTIPLIER // ANN

        # 2*N*K0 / _g1k0
        mul2: int = (2 * 10**18) * n_coins * K0 // _g1k0

        neg_fprime: int = (
            (S + S * mul2 // 10**18) + mul1 * n_coins // K0 - mul2 * D // 10**18
        )

        # D -= f / fprime
        D_plus: int = D * (neg_fprime + S) // neg_fprime
        D_minus: int = D * D // neg_fprime
        if 10**18 > K0:
            D_minus += D * (mul1 // neg_fprime) // 10**18 * (10**18 - K0) // K0
        else:
            D_minus -= D * (mul1 // neg_fprime) // 10**18 * (K0 - 10**18) // K0

        if D_plus > D_minus:
            D = D_plus - D_minus
        else:
            D = (D_minus - D_plus) // 2

        diff = abs(D - D_prev)
        # Could reduce precision for gas efficiency here
        if diff * 10**14 < max(10**16, D):
            # Test that we are safe with the next newton_y
            for _x in x:
                frac: int = _x * 10**18 // D
                if frac < 10**16 or frac > 10**20:
                    raise CalculationError("Unsafe value for x[i]")
            return int(D)

    raise CalculationError("Did not converge")



from numpy import isnan
from scipy.optimize import least_squares

from curvesim.logging import get_logger
from curvesim.templates.trader import Trade, Trader

from ..common import get_arb_trades

logger = get_logger(__name__)


class VolumeLimitedArbitrageur(Trader):
    """
    Computes, executes, and reports out arbitrage trades.
    """

    def compute_trades(self, prices, volume_limits):  # pylint: disable=arguments-differ
        """
        Computes trades to optimally arbitrage the pool, constrained by volume limits.

        Parameters
        ----------
        prices : dict
            Current market prices from the price_sampler.

        volume_limits : dict
            Current volume limits for each trading pair.


        Returns
        -------
        trades : list of :class:`Trade` objects
            List of trades to perform.

        additional_data: dict
            Dict of additional data to be passed to the state log as part of trade_data.
        """

        trades, errors, _ = multipair_optimal_arbitrage(
            self.pool, prices, volume_limits
        )
        return trades, {"price_errors": errors}


def multipair_optimal_arbitrage(  # noqa: C901  pylint: disable=too-many-locals
    pool, prices, limits
):
    """
    Computes trades to optimally arbitrage the pool, constrained by volume limits.

    Parameters
    ----------
    pool :
        Simulation interface to a subclass of :class:`Pool`.

    prices : dict
        Current market prices from the price_sampler.

    volume_limits : dict
        Current volume limits for each trading pair.

    Returns
    -------
    trades : List[Tuple]
        List of trades to perform.
        Trades are formatted as (coin_i, coin_j, trade_size)

    errors : List[Float]
        Post-trade price error between pool price and market price for each token pair.

    res : scipy.optimize.OptimizeResult
        Results object from the numerical optimizer.
    """
    init_trades = get_arb_trades(pool, prices)

    # Limit trade size, add size bounds
    limited_init_trades = []
    for t in init_trades:
        size, pair, price_target = t
        limit = int(limits[pair] * 10**18)
        t = min(size, limit), pair, price_target, 0, limit + 1
        limited_init_trades.append(t)

    # Order trades in terms of expected size
    limited_init_trades = sorted(limited_init_trades, reverse=True, key=lambda t: t[0])
    sizes, coins, price_targets, lo, hi = zip(*limited_init_trades)

    def post_trade_price_error_multi(dxs, price_targets, coins):
        with pool.use_snapshot_context():
            for k, pair in enumerate(coins):
                if isnan(dxs[k]):
                    dx = 0
                else:
                    dx = int(dxs[k])

                coin_in, coin_out = pair
                min_size = pool.get_min_trade_size(coin_in)
                if dx > min_size:
                    pool.trade(coin_in, coin_out, dx)

            errors = []
            for k, pair in enumerate(coins):
                price = pool.price(*pair, use_fee=True)
                errors.append(price - price_targets[k])

        return errors

    # Find trades that minimize difference between
    # pool price and external market price
    trades = []
    try:
        res = least_squares(
            post_trade_price_error_multi,
            x0=sizes,
            args=(price_targets, coins),
            bounds=(lo, hi),
            gtol=10**-15,
            xtol=10**-15,
        )

        # Format trades into tuples, ignore if dx=0
        dxs = res.x

        for k, amount_in in enumerate(dxs):
            if isnan(amount_in):
                continue

            amount_in = int(amount_in)
            coin_in, coin_out = coins[k]
            min_size = pool.get_min_trade_size(coin_in)
            if amount_in > min_size:
                trades.append(Trade(coin_in, coin_out, amount_in))

        errors = res.fun

    except Exception:
        logger.error(
            "Optarbs args: x0: %s, lo: %s, hi: %s, prices: %s",
            sizes,
            lo,
            hi,
            price_targets,
            exc_info=True,
        )
        errors = post_trade_price_error_multi([0] * len(sizes), price_targets, coins)
        res = []

    return trades, errors, res



import pytest

from curvesim.exceptions import SubgraphError
from curvesim.network.subgraph import _volume, pool_snapshot
from curvesim.network.utils import sync

ZERO_ADDRESS = "0x" + "00" * 20


def test_convex_subgraph_volume_query():
    """Test the volume query."""

    chain = "mainnet"
    address = "0xbEbc44782C7dB0a1A60Cb6fe97d0b483032FF1C7"
    _volume_sync = sync(_volume)
    volumes = _volume_sync(address, chain, env="prod", days=2, end=1686700800)
    assert len(volumes) == 2

    volumes = _volume_sync(ZERO_ADDRESS, chain, env="prod", days=2, end=1686700800)
    assert len(volumes) == 0


def test_convex_subgraph_stableswap_snapshot_query():
    """Test the pool snapshot query for stableswap."""

    chain = "mainnet"
    address = "0xbEbc44782C7dB0a1A60Cb6fe97d0b483032FF1C7"
    end_ts = 1687305600
    _snapshot_sync = sync(pool_snapshot)
    snapshot = _snapshot_sync(address, chain, end_ts=end_ts)
    assert snapshot["address"] == address
    assert snapshot["chain"] == chain

    assert snapshot["timestamp"] <= end_ts

    assert snapshot["version"] == 1

    params = snapshot["params"]
    assert params["A"] > 0

    with pytest.raises(SubgraphError):
        _snapshot_sync(ZERO_ADDRESS, chain)


def test_convex_subgraph_cryptoswap_snapshot_query():
    """Test the pool snapshot query for cryptoswap."""

    chain = "mainnet"
    address = "0x3211C6cBeF1429da3D0d58494938299C92Ad5860"
    end_ts = 1687305600
    _snapshot_sync = sync(pool_snapshot)
    snapshot = _snapshot_sync(address, chain, end_ts=end_ts)
    assert snapshot["address"] == address
    assert snapshot["chain"] == chain

    assert snapshot["timestamp"] <= end_ts

    assert snapshot["version"] == 2

    params = snapshot["params"]
    assert params["A"] > 0
    assert params["gamma"] > 0



"""
This module provides functionality for creating Altair charts with custom styles.

It contains utility functions for creating charts and default chart properties. It
uses styles defined in the .styles module and allows custom configuration of chart
properties.
"""
from altair import Chart, Scale, data_transformers

from curvesim.exceptions import PlotError

from .chart_properties import make_chart_kwargs
from .styles import STYLES

data_transformers.disable_max_rows()


def make_chart(config, x=None, y=None, color=None, **kwargs):
    """
    Create an interactive chart with custom properties and styles.

    Parameters
    ----------
    config : dict
        The configuration dictionary for the chart. Must contain a 'style' key.
    x : str or dict, optional
        The x-axis property. Can be a shorthand string or a dictionary
        of alt.X kwargs.
    y : str or dict, optional
        The y-axis property. Can be a shorthand string or a dictionary
        of alt.Y kwargs.
    color : str or dict, optional
        The color property. Can be a shorthand string or a dictionary
        of alt.Color kwargs.
    **kwargs
        Additional keyword arguments are added to the chart configuration.

    Returns
    -------
    altair.Chart
        The created chart.

    Raises
    ------
    PlotError
        If no style is found in the config or if the style is not found in STYLES.
    """
    title = config.get("title", "")

    try:
        style = config["style"]
    except KeyError as e:
        raise PlotError(f"No style found in {title} config.") from e

    try:
        style = STYLES[style]
    except KeyError as e:
        raise PlotError(f"Style {style} not found in plot.styles.") from e

    defaults = make_defaults(title, x, y, color)
    chart_kwargs = make_chart_kwargs(defaults, [style, config, kwargs])
    return Chart(**chart_kwargs).interactive()


def make_defaults(title, x=None, y=None, color=None):
    """
    Create default properties for a chart.

    Parameters
    ----------
    title : str
        The title of the chart.
    x : str or dict, optional
        The x-axis property. Can be a shorthand string or a dictionary
        of alt.X kwargs.
    y : str or dict, optional
        The y-axis property. Can be a shorthand string or a dictionary
        of alt.Y kwargs.
    color : str or dict, optional
        The color property. Can be a shorthand string or a dictionary
        of alt.Color kwargs.

    Returns
    -------
    dict
        The default properties for a chart.

    Raises
    ------
    PlotError
        If x, y, or color is not a string (shorthand) or
        a dictionary (altair kwargs).
    """
    defaults = {
        "title": title,
        "encoding": {
            "x": {},
            "y": {"title": title},
            "color": {"scale": Scale(scheme="viridis")},
        },
    }

    for key, val in zip(["x", "y", "color"], [x, y, color]):
        if val:
            if isinstance(val, str):
                val = {"shorthand": val}

            elif not isinstance(val, dict):
                cap = key.capitalize()
                raise PlotError(
                    f"{key} must be a string (shorthand) or dict (alt.{cap} kwargs)."
                )

            for k, v in val.items():
                defaults["encoding"][key][k] = v

    return defaults



from math import prod

import pytest
from hypothesis import HealthCheck, given, settings
from hypothesis import strategies as st

from curvesim.exceptions import ParameterSamplerError
from curvesim.iterators.param_samplers import ParameterizedPoolIterator
from curvesim.iterators.param_samplers.parameterized_pool_iterator import (
    DEFAULT_POOL_MAP,
)
from curvesim.pool.cryptoswap.calcs import newton_D
from curvesim.pool.sim_interface import SimCurveCryptoPool

# Strategies
POOL_PARAMS = {
    "A": st.integers(min_value=1, max_value=3000),
    "D": st.integers(min_value=10**24, max_value=10**28),
    "fee": st.integers(min_value=10**10, max_value=10**11),
}

METAPOOL_PARAMS = {
    "A": st.integers(min_value=1, max_value=3000),
    "D": st.integers(min_value=10**24, max_value=10**28),
    "fee": st.integers(min_value=10**10, max_value=10**11),
    "A_base": st.integers(min_value=10, max_value=3000),
    "D_base": st.integers(min_value=10**24, max_value=10**28),
    "fee_base": st.integers(min_value=10**10, max_value=10**11),
}

CRYPTOPOOL_PARAMS = {
    "D": st.integers(min_value=10**24, max_value=10**28),
    "A": st.integers(min_value=4000, max_value=4000000000),
    "gamma": st.integers(min_value=10**10, max_value=2 * 10**16),
    "mid_fee": st.integers(min_value=5 * 10**5, max_value=10**10),
    "out_fee": st.integers(min_value=5 * 10**5, max_value=10**10),
    "allowed_extra_profit": st.integers(min_value=0, max_value=10**8),
    "fee_gamma": st.integers(min_value=0, max_value=10**10),
    "adjustment_step": st.integers(min_value=0, max_value=10**10),
    "ma_half_time": st.integers(min_value=0, max_value=604800),
}


def make_parameter_strats(parameters):
    """
    Returns param_subset_strats for variable_params and fixed_params arguments.

    Note: variable params are limited to length 1 and fixed params limited to length 0.
    Variable params of length 0 are tested seperately in
    test_ParameterizedPoolIterator_no_variable_params.
    """
    return (
        param_subset_strat(parameters, val_to_list=True),
        param_subset_strat(parameters, min_size=0),
    )


@st.composite
def param_subset_strat(draw, parameters, min_size=1, max_size=None, val_to_list=False):
    """
    Generates a dict including a subset of the provided parameters. This simulates
    user inputs to the parameter sampler (i.e., variable_params, fixed_params).

    Parameter values are taken from param_strats. If val_to_list is True, generates a
    list of values for each parameter (i.e., in the format of variable_params).
    """
    max_size = max_size or len(parameters)

    key_strat = to_list_strat(
        st.sampled_from(list(parameters)), min_size=min_size, max_size=max_size
    )

    key_subset = draw(key_strat)
    dict_subset = {}
    for key in key_subset:
        val_strat = parameters[key]
        if val_to_list:
            val_strat = to_list_strat(val_strat, max_size=3)
        dict_subset[key] = draw(val_strat)

    return dict_subset


def to_list_strat(strategy, min_size=2, max_size=20):
    """Makes list strategy from input strategy"""
    return st.lists(strategy, min_size=min_size, max_size=max_size, unique=True)


# Tests
def test_invalid_parameter_exceptions(sim_curve_pool):
    """
    Test exceptions for invalid parameter names. If fails,
    ParameterSampler._validate_attributes is malfunctioning.
    """

    # Not a pool param
    with pytest.raises(ParameterSamplerError):
        ParameterizedPoolIterator(sim_curve_pool, {"not_a_param": [20, 30]})

    # Basepool param when no basepool
    with pytest.raises(ParameterSamplerError):
        ParameterizedPoolIterator(sim_curve_pool, {"A_base": [10, 100]})


def test_ParameterizedPoolIterator_subclass_mapping():
    """
    Test ParameterizedPoolIterator subclass mapping and instantiation.
    If fails, __new__ or subclass instantiation is failing.
    """

    class DummyPool:
        def __init__(self):
            for attr in "abcd":
                setattr(self, attr, None)

    class DummyParamSampler(ParameterizedPoolIterator):
        @property
        def _pool_type(self):
            return DummyPool

    mapping = {DummyPool: DummyParamSampler}
    pool = DummyPool()
    variable_params = {"a": [1, 2], "b": [3, 4]}
    fixed_params = {"c": 5, "d": 6}

    param_sampler = ParameterizedPoolIterator(
        pool, variable_params, fixed_params, pool_map=mapping
    )

    assert isinstance(param_sampler, DummyParamSampler)
    assert isinstance(param_sampler.pool_template, DummyPool)


def test_ParameterizedPoolIterator_unmapped_pool_exception():
    """
    Test ParameterizedPoolIterator exception for unmapped pool type.
    If fails, ParameterizedPoolIterator.__new__ is failing.
    """

    class DummyPool:
        def __init__(self):
            for attr in "ab":
                setattr(self, attr, None)

    pool = DummyPool()

    with pytest.raises(ParameterSamplerError):
        ParameterizedPoolIterator(pool)


def test_ParameterizedPoolIterator_wrong_pool_exception():
    """
    Test subclass exceptions when instantiated with wrong pool type. If fails,
    ._validate_pool_type or a subclass's _pool_type property is malfunctioning.
    """

    class DummyPool:
        pass

    pool = DummyPool()

    for subclass in DEFAULT_POOL_MAP.values():
        mapping = {DummyPool: subclass}

        with pytest.raises(ParameterSamplerError):
            ParameterizedPoolIterator(pool, pool_map=mapping)


def test_ParameterizedPoolIterator_pool_template_mutation():
    """
    Test that parameter iteration doesn't mutate the pool template.
    """

    class DummyPool:
        def __init__(self):
            self.a = 0
            self.b = 0

    class DummyParamSampler(ParameterizedPoolIterator):
        @property
        def _pool_type(self):
            return DummyPool

    pool = DummyPool()
    variable_params = {"a": [1, 2], "b": [3, 4]}
    mapping = {DummyPool: DummyParamSampler}

    param_sampler = ParameterizedPoolIterator(pool, variable_params, pool_map=mapping)

    for pool_sample, pool_params in param_sampler:
        assert param_sampler.pool_template.a == 0
        assert param_sampler.pool_template.b == 0


def test_ParameterizedPoolIterator_no_variable_params():
    """
    Test that .parameter_sequence uses fixed params when variable params are omitted.
    """

    class DummyPool:
        def __init__(self):
            self.a = 0
            self.b = 0

    class DummyParamSampler(ParameterizedPoolIterator):
        @property
        def _pool_type(self):
            return DummyPool

    pool = DummyPool()
    mapping = {DummyPool: DummyParamSampler}
    fixed_params = {"a": 1, "b": 2}

    param_sampler = ParameterizedPoolIterator(
        pool, fixed_params=fixed_params, pool_map=mapping
    )

    assert param_sampler.parameter_sequence == [fixed_params]


@given(*make_parameter_strats(POOL_PARAMS))
@settings(
    max_examples=5,
    deadline=None,
    suppress_health_check=[HealthCheck.function_scoped_fixture],
)
def test_ParameterizedPoolIterator_curve_pool(
    sim_curve_pool, variable_params, fixed_params
):
    """Tests full instantiation of ParameterizedCurvePoolIterator from SimCurvePool."""
    _test_ParameterizedPoolIterator(sim_curve_pool, variable_params, fixed_params)


@given(*make_parameter_strats(METAPOOL_PARAMS))
@settings(
    max_examples=5,
    deadline=None,
    suppress_health_check=[HealthCheck.function_scoped_fixture],
)
def test_ParameterizedPoolIterator_curve_meta_pool(
    sim_curve_meta_pool, variable_params, fixed_params
):
    """Tests full instantiation of ParameterizedCurveMetaPoolIterator from SimCurveMetaPool."""

    def fixed_virtual_price():
        return 10**18

    sim_curve_meta_pool.basepool.get_virtual_price = fixed_virtual_price

    _test_ParameterizedPoolIterator(sim_curve_meta_pool, variable_params, fixed_params)


@given(*make_parameter_strats(METAPOOL_PARAMS))
@settings(
    max_examples=5,
    deadline=None,
    suppress_health_check=[HealthCheck.function_scoped_fixture],
)
def test_ParameterizedPoolIterator_curve_rai_pool(
    sim_curve_rai_pool, variable_params, fixed_params
):
    """Tests full instantiation of ParameterizedCurveMetaPoolIterator from SimCurveRaiPool."""

    def fixed_virtual_price():
        return 10**18

    sim_curve_rai_pool.basepool.get_virtual_price = fixed_virtual_price

    _test_ParameterizedPoolIterator(sim_curve_rai_pool, variable_params, fixed_params)


@given(*make_parameter_strats(CRYPTOPOOL_PARAMS))
@settings(
    max_examples=5,
    deadline=None,
    suppress_health_check=[HealthCheck.function_scoped_fixture],
)
def test_ParameterizedPoolIterator_curve_crypto_pool(
    sim_curve_crypto_pool, variable_params, fixed_params
):
    """Tests full instantiation of ParameterizedCurveCryptoPoolIteratorfrom SimCurveCryptoPool."""
    _test_ParameterizedPoolIterator(
        sim_curve_crypto_pool, variable_params, fixed_params
    )


# Helper functions for tests
def _test_ParameterizedPoolIterator(pool, variable_params, fixed_params, pool_map=None):
    """
    Tests correct instantiation of a ParameterizedPoolIterator subclass, including
    correct class and parameters/attributes.
    """
    param_sampler = ParameterizedPoolIterator(
        pool, variable_params, fixed_params, pool_map
    )

    pool_to_sampler = pool_map or DEFAULT_POOL_MAP
    expected_type = pool_to_sampler[type(pool)]

    assert isinstance(param_sampler, expected_type)

    _test_pool_params(param_sampler.pool_template, fixed_params)
    _test_ParameterizedPoolIterator_parameter_sequence(param_sampler, variable_params)
    _test_ParameterizedPoolIterator_iterations(param_sampler)


def _test_pool_params(pool, params):
    """
    Tests that pool attributes match those defined in the params dict.
    """
    for key, val in params.items():
        _pool, _key = parse_pool_attribute(pool, key)

        if _key == "D":
            if isinstance(pool, SimCurveCryptoPool):
                D_from_xp = newton_D(_pool.A, _pool.gamma, _pool._xp())
                assert abs(_pool.D - val) <= 10**10
                assert (
                    abs(D_from_xp - val) <= 10**10
                )  # _newton_D allows some minor error

            else:
                D = _pool.D()
                assert abs(D - val) <= 3

        else:
            assert getattr(_pool, _key) == val


def _test_ParameterizedPoolIterator_parameter_sequence(param_sampler, variable_params):
    """
    Tests that param_sampler.parameter_sequence has the expected properties for the
    product of the input variable params.
    """

    param_sequence = param_sampler.parameter_sequence
    param_keys, param_values = zip(*variable_params.items())

    # Test expected number of combinations
    lengths = [len(vals) for vals in param_values]
    length_product = prod(lengths)
    assert len(param_sequence) == length_product

    # Test all keys present in each dict
    input_keys = variable_params.keys()
    for params in param_sequence:
        assert params.keys() == input_keys  # note: undordered equality

    # Test no duplicate entries
    for params1 in param_sequence:
        same_entry_count = 0
        for params2 in param_sequence:
            if params1 == params2:
                same_entry_count += 1
        assert same_entry_count == 1

    # Test expected count of each parameter value
    all_param_values = {key: [dct[key] for dct in param_sequence] for key in param_keys}
    for key, values in variable_params.items():
        param_list = all_param_values[key]
        expected_repeats = length_product / len(values)
        for val in values:
            assert param_list.count(val) == expected_repeats


def _test_ParameterizedPoolIterator_iterations(param_sampler):
    """
    Tests that .parameter_sequence is applied to pool on each iter.
    """
    sequences = zip(param_sampler, param_sampler.parameter_sequence)

    for (pool_sample, pool_params), sampler_params in sequences:
        assert pool_params == sampler_params  # test if correct vals
        _test_pool_params(pool_sample, pool_params)  # test if vals applied to pool


def parse_pool_attribute(pool, attribute):
    """
    Helper function to route "_base" attributes to basepool if necessary.
    """
    if attribute.endswith("_base"):
        return pool.basepool, attribute[:-5]

    return pool, attribute



from scipy.optimize import root_scalar

from curvesim.logging import get_logger
from curvesim.metrics import metrics as Metrics

logger = get_logger(__name__)
DEFAULT_METRICS = [
    Metrics.Timestamp,
    Metrics.PoolValue,
    Metrics.PoolBalance,
    # Metrics.PriceDepth,
    Metrics.PoolVolume,
    Metrics.ArbMetrics,
]

DEFAULT_PARAMS = {
    "A": [int(2 ** (a / 2)) for a in range(12, 28)],
    "fee": list(range(1000000, 5000000, 1000000)),
}

TEST_PARAMS = {"A": [100, 1000], "fee": [3000000, 4000000]}
TEST_CRYPTO_PARAMS = {
    "A": [270000, 2700000],
    "gamma": [1300000000000, 13000000000],
    "fee_gamma": [500000000000000, 50000000000000],
    "out_fee": [80000000, 800000000],
}


def get_arb_trades(pool, prices):
    """
    Returns triples of "trades", one for each coin pair in `combo`.

    Each trade is a triple consisting of size, ordered coin-pair,
    and price target to move the pool price to.

    Parameters
    ----------
    pool: SimPool
        Pool to arbitrage on

    prices : iterable
        External market prices for each coin-pair


    Returns
    -------
    trades: List[Tuple]
        List of triples (size, coins, price_target)
        "size": trade size
        "coins": in token, out token
        "price_target": price target for arbing the token pair
    """

    def post_trade_price_error(dx, coin_in, coin_out, price_target):
        with pool.use_snapshot_context():
            dx = int(dx)
            if dx > 0:
                pool.trade(coin_in, coin_out, dx)
            price = pool.price(coin_in, coin_out, use_fee=True)

        return price - price_target

    trades = []

    for pair in prices:
        i, j = pair

        if pool.price(i, j) - prices[pair] > 0:
            price = prices[pair]
            coin_in, coin_out = i, j
        elif pool.price(j, i) - 1 / prices[pair] > 0:
            price = 1 / prices[pair]
            coin_in, coin_out = j, i
        else:
            trades.append((0, pair, prices[pair]))
            continue

        high = pool.get_max_trade_size(coin_in, coin_out)
        bounds = (0, high)
        try:
            res = root_scalar(
                post_trade_price_error,
                args=(coin_in, coin_out, price),
                bracket=bounds,
                method="brentq",
            )
            size = int(res.root)
        except ValueError:
            pool_price = pool.price(coin_in, coin_out)
            logger.error(
                "Opt_arb error: Pair: (%s, %s), Pool price: %s,"
                "Target Price: %s, Diff: %s",
                coin_in,
                coin_out,
                pool_price,
                price,
                pool_price - price,
            )
            size = 0

        trades.append((size, (coin_in, coin_out), price))

    return trades



from curvesim.logging import get_logger
from curvesim.price_data import get
from curvesim.templates.price_samplers import PriceSample, PriceSampler
from curvesim.utils import dataclass, override

logger = get_logger(__name__)


@dataclass(slots=True)
class PriceVolumeSample(PriceSample):
    """
    Attributes
    -----------
    timestamp : datetime.datetime
        Timestamp for the current price/volume.
    prices : dict
        Price for each pairwise coin combination.
    volumes : dict
        Volume for each pairwise coin combination.
    """

    volumes: dict


class PriceVolume(PriceSampler):
    """
    An iterator that retrieves price/volume and iterates over timepoints in the data.
    """

    def __init__(
        self,
        assets,
        *,
        days=60,
        data_dir="data",
        src="coingecko",
        end=None,
    ):
        """
        Retrieves price/volume data and prepares it for iteration.

        Parameters
        ----------
        assets: SimAssets
            Object giving the properties of the assets for simulation
            (e.g., symbols, addresses, chain)

        days: int, defaults to 60
            Number of days to pull data for.

        data_dir: str, defaults to "data"
            Relative path to saved data folder.

        src: str, defaults to "coingecko"
            Identifies pricing source: coingecko or local.

        """
        prices, volumes, _ = get(
            assets.addresses,
            chain=assets.chain,
            days=days,
            data_dir=data_dir,
            src=src,
            end=end,
        )

        self.prices = prices.set_axis(assets.symbol_pairs, axis="columns")
        self.volumes = volumes.set_axis(assets.symbol_pairs, axis="columns")

    @override
    def __iter__(self) -> PriceVolumeSample:
        """
        Yields
        -------
        :class:`PriceVolumeSample`
        """
        for price_row, volume_row in zip(
            self.prices.iterrows(), self.volumes.iterrows()
        ):
            price_timestamp, prices = price_row
            volume_timestamp, volumes = volume_row
            assert (
                price_timestamp == volume_timestamp
            ), "Price/volume timestamps don't match"

            prices = prices.to_dict()
            volumes = volumes.to_dict()

            yield PriceVolumeSample(price_timestamp, prices, volumes)

    def total_volumes(self):
        """
        Returns
        -------
        pandas.Series
            Total volume for each pairwise coin combination, summed accross timestamps.
        """
        return self.volumes.sum().to_dict()
